#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <error.h>
#include <errno.h>

#include <unistd.h>
#include <fcntl.h>

#include <locale.h>
#include <libintl.h>

#include <expat.h>

#define _(string) gettext (string)


const char datafile[]="/usr/share/X11/xkb/rules/xorg.xml";
const char po_domain[]="xkeyboard-config";

#define MAXLINE 1024

typedef struct user_data_struct {
    char path[MAXLINE];

    FILE *file;

    char name[MAXLINE];
    char description[MAXLINE];

    char layname[MAXLINE];
    char laydescription[MAXLINE];
} user_data;

void
append (char *str, const char *buf, int len)
{
    int str_len = strlen(str);

    if ( len < MAXLINE - (str_len + 1) ) {
        strncpy (&str[str_len], buf, len);
        str[str_len + len] = '\0';
    } else if (MAXLINE - (str_len + 1) > 0) {
        strncpy (&str[str_len], buf, MAXLINE - (str_len + 1));
        str[MAXLINE] = '\0';
    } else
        fprintf (stderr, "WARNING: CDATA is too long!\n");
}

void
xfopen (FILE **file, const char *name)
{
    if (!*file)
        *file = fopen (name, "w");
    else
        error (EXIT_FAILURE, errno,
               "ERROR: Already open");
    if (!*file)
        error (EXIT_FAILURE, errno,
               "ERROR: Unable to open file %s for writing", name);
}

void
reset (user_data *u, int keep_layout)
{
    if (!keep_layout) {
        u->layname[0] = '\0';
        u->laydescription[0] = '\0';
    }
    u->name[0] = '\0';
    u->description[0] = '\0';
}

void
trim (user_data *u)
{
    int pos = strlen (u->path) - 1;
    while (pos > 0 && u->path[pos] != '/') pos--;
    u->path[pos] = '\0';
}

static void XMLCALL
element_start (void *data, const char *el, const char **attr)
{
    user_data *u = data;

    append (u->path, "/", 1);
    append (u->path, el, strlen (el));
    
    if (0 == strcmp (u->path, "/xkbConfigRegistry/layoutList"))
        xfopen (&u->file, "./layouts");
    else if (0 == strcmp (u->path, "/xkbConfigRegistry/modelList"))
        xfopen (&u->file, "./models");
    else if (0 == strcmp (u->path, "/xkbConfigRegistry/optionList"))
        xfopen (&u->file, "./options");
}
    
static void XMLCALL
element_stop (void *data, const char *el)
{
    user_data *u = data;

    if (0 == strcmp (u->path, "/xkbConfigRegistry/layoutList/layout/configItem")) {
        fprintf (u->file,"%s\t%s\n",
                 u->layname,
                 _(u->laydescription));
        reset (u, 1);
    } else if (0 == strcmp (u->path, "/xkbConfigRegistry/layoutList/layout/variantList/variant/configItem")) {
        fprintf (u->file,"%s(%s)\t%s - %s\n",
                 u->layname,
                 u->name,
                 _(u->laydescription),
                 _(u->description));
        reset (u, 1);
    } else if (0 == strcmp (u->path, "/xkbConfigRegistry/layoutList/layout")) {
        reset (u, 0);
    } else if (0 == strcmp (u->path, "/xkbConfigRegistry/modelList/model/configItem")) {
        fprintf (u->file,"%s\t%s\n",
                 u->name,
                 _(u->description));
        reset (u, 0);
    } else if (0 == strcmp (u->path, "/xkbConfigRegistry/optionList/group/option/configItem")) {
        fprintf (u->file,"%s\t%s\n",
                 u->name,
                 _(u->description));
        reset (u, 0);
    }

    trim (u);

    if (0 == strcmp (u->path, "/xkbConfigRegistry") && u->file) {
        fclose (u->file);
        u->file = NULL;
    }
}

static void XMLCALL
cdata_handler (void *data, const char *buffer, int len)
{
    user_data *u = data;

    if (0 == strcmp (u->path, "/xkbConfigRegistry/layoutList/layout/configItem/name"))
        append (u->layname, buffer, len);
    else if (0 == strcmp (u->path, "/xkbConfigRegistry/layoutList/layout/configItem/description"))
        append (u->laydescription, buffer, len);
    else if (0 == strcmp (u->path, "/xkbConfigRegistry/layoutList/layout/variantList/variant/configItem/name") ||
             0 == strcmp (u->path, "/xkbConfigRegistry/modelList/model/configItem/name") ||
             0 == strcmp (u->path, "/xkbConfigRegistry/optionList/group/option/configItem/name"))
        append (u->name, buffer, len);
    else if (0 == strcmp (u->path, "/xkbConfigRegistry/layoutList/layout/variantList/variant/configItem/description") ||
               0 == strcmp (u->path, "/xkbConfigRegistry/modelList/model/configItem/description") ||
               0 == strcmp (u->path, "/xkbConfigRegistry/optionList/group/option/configItem/description"))
        append (u->description, buffer, len);
}

int 
main (int argc,char **argv)
{
    int fd;
    char buffer[BUFSIZ];
    size_t len;
    XML_Parser parser;
    user_data data;
    int ret = EXIT_SUCCESS;

    setlocale (LC_ALL,"");
    textdomain (po_domain);
    bindtextdomain (po_domain, "/usr/share/locale");

    if (!(parser = XML_ParserCreate (NULL)))
    error (EXIT_FAILURE, 0, "couldn't allocate memory for parser");

    if ((fd = open (datafile, O_RDONLY)) < 0)
    error (EXIT_FAILURE, errno, "unable to open database file %s", datafile);

    XML_SetElementHandler(parser, element_start, element_stop);
    XML_SetCharacterDataHandler(parser, cdata_handler);

    memset (&data, 0, sizeof (data));
    XML_SetUserData (parser, &data);

    while ((len = read (fd, buffer, (sizeof buffer) - 1)) > 0)
    {
        if (XML_Parse (parser, buffer, len, XML_FALSE) == XML_STATUS_ERROR)
        {
            fprintf (stderr,"Parse error at line %lu:%s",
                     XML_GetCurrentLineNumber (parser),
                     XML_ErrorString(XML_GetErrorCode (parser)));
            ret = EXIT_FAILURE;
            goto exit;
        }
    }

    XML_Parse (parser, buffer, 0, XML_TRUE);
    
 exit:
    XML_ParserFree (parser);

    return ret;
}
